<?php

namespace Drupal\tfa;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Plugin\DefaultPluginManager;

/**
 * Tfa plugin manager.
 *
 * @internal
 */
class TfaPluginManager extends DefaultPluginManager {

  /**
   * TFA configuration object.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $tfaSettings;

  /**
   * Constructs TfaPluginManager object.
   *
   * @param \Traversable $namespaces
   *   An object that implements \Traversable which contains the root paths
   *   keyed by the corresponding namespace to look for plugin implementations.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache_backend
   *   Cache backend instance to use.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $module_handler
   *   The module handler to invoke the alter hook with.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The config factory.
   */
  public function __construct(\Traversable $namespaces, CacheBackendInterface $cache_backend, ModuleHandlerInterface $module_handler, ConfigFactoryInterface $config_factory) {
    parent::__construct(
      'Plugin/Tfa',
      $namespaces,
      $module_handler,
      'Drupal\tfa\TfaPluginInterface',
      'Drupal\tfa\Annotation\Tfa'
    );
    $this->alterInfo('tfa_info');
    $this->setCacheBackend($cache_backend, 'tfa_plugins');
    $this->tfaSettings = $config_factory->get('tfa.settings');
  }

  /**
   * Return plugin definitions for all validation plugins.
   *
   * @return array
   *   Array of validation plugin definitions.
   */
  public function getValidationDefinitions(): array {
    return $this->getClassDefinitions('\Drupal\tfa\Plugin\TfaValidationInterface');
  }

  /**
   * Return plugin definitions for all login plugins.
   *
   * @param bool $active_only
   *   Whether to only get active plugins enabled in config.
   *
   * @return array
   *   Array of login plugin definitions.
   */
  public function getLoginDefinitions(bool $active_only = TRUE): array {
    $plugins = $this->getClassDefinitions('\Drupal\tfa\Plugin\TfaLoginInterface');
    if ($active_only) {
      $login_plugins = $this->tfaSettings->get('login_plugins');
      $plugins = array_filter($plugins, function (array $plugin) use ($login_plugins) {
        return ($login_plugins[$plugin['id']] ?? NULL) === $plugin['id'];
      });
    }
    return $plugins;
  }

  /**
   * Return plugin definitions for all send plugins.
   *
   * @return array
   *   Array of send plugin definitions.
   */
  public function getSendDefinitions(): array {
    return $this->getClassDefinitions('\Drupal\tfa\Plugin\TfaSendInterface');
  }

  /**
   * Return plugin definitions for all plugins implementing the given class.
   *
   * @param string $class
   *   Name of the class or interface that the plugins must implement.
   *
   * @return array
   *   Array of plugin definitions.
   */
  public function getClassDefinitions(string $class): array {
    $all_plugins = $this->getDefinitions();
    $plugins = [];

    foreach ($all_plugins as $key => $plugin) {
      if (is_a($plugin['class'], $class, TRUE)) {
        $plugins[$key] = $plugin;
      }
    }
    return $plugins;
  }

}
